from functools import lru_cache
from typing import TYPE_CHECKING, Dict, Union, Tuple
from dataclasses import dataclass
from pydash import snake_case

from sedaro_base_client.api_client import Api
from .settings import UPDATE, DELETE
from .exceptions import NonexistantBlockError
from .utils import temp_crud

if TYPE_CHECKING:
    from .block_class_client import BlockClassClient
    from .sedaro_api_client import SedaroApiClient
    from .branch_client import BranchClient


@dataclass
class BlockClient:
    id: str
    _block_class_client: 'BlockClassClient'
    '''Class for interacting with all Blocks of this class type'''

    def __str__(self) -> str:
        attrs = ''
        for k, v in self.data.items():
            if type(v) is str:
                # TODO: this is often not True when it should be b/c so many values become the `DynamicSchema` class
                # Figure out how to fix this
                v = f"'{v}'"
            attrs += f'\n   {k}={v}'
        return f'\n{self._block_name}({attrs}\n)\n'

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.id == other.id

    def __hash__(self):
        # allows a BlockClient to be a key in a dict and @lru_cache wrapper to work on methods on this class
        return hash(self.__class__.__name__ + self.id)

    def __getattr__(self, key: str) -> any:
        """Allows for dotting into the `BlockClient` to access keys on the referenced Sedaro Block. Additionally, makes
        it so dotting into relationship fields returns `BlockClient`s corresponding to the related Sedaro Blocks.

        Args:
            key (str): attribute being keyed into

        Raises:
            AttributeError: if the attribute doesn't exist on the refrenced Sedaro Block

        Returns:
            any: the value of the corresponding attribute on the referenced Sedaro Block
        """
        if key not in self.data:
            raise make_attr_error(key, self._block_name)

        val = self.data[key]

        if not self.is_rel_field(key) or val in (None, [], {}):
            return val

        side_type = self.get_rel_field_type(key)
        branch_client = self._branch_client

        if side_type == MANY_SIDE:
            return [branch_client.get_block_client(id) for id in val]

        if side_type == MANY_SIDE_DATA:
            return {branch_client.get_block_client(id): data for id, data in val.items()}

        if side_type == ONE_SIDE:
            return branch_client.get_block_client(val)

        # TODO: wasn't able to use this below, because the types are sometimes of DynamicSchema and not the actual
        # primitive types... Need to figure this out.
        # if isinstance(val, list):
        #     return [branch_client.get_block_client(id) for id in val]

        # if isinstance(val, dict):
        #     return {branch_client.get_block_client(id): data for id, data in val.items()}

        # if isinstance(val, str):
        #     return branch_client.get_block_client(val)

        raise NotImplementedError(
            f'Unsupported relationship type on "{self._block_name}", attribute: "{key}".'
        )

    @property
    def data(self) -> Dict:
        '''The properties of the corresponding Sedaro Block as a dictionary'''
        self.enforce_still_exists()
        return self._branch_client.data[self._block_group][self.id]

    @property
    def _block_name(self) -> str:
        '''The name of the Sedaro Block class associated with this `BlockClient`'''
        try:
            # NOTE: can't use `check_still_exists` instead of try/except, b/c that method calls this one
            return self._branch_client._block_id_to_type_map[self.id]
        except KeyError:
            # this won't be as specific, but allows to know at least the parent type even after a block was deleted
            return self._block_class_client._block_name

    @property
    def _block_group(self) -> str:
        '''The name of the Sedaro Block Group this type of `BlockClient`'s associated Sedaro Block is stored in'''
        return self._branch_client._block_class_to_block_group_map[self._block_name]

    @property
    def _branch_client(self) -> 'BranchClient':
        '''The `BranchClient` this `BlockClient` is connected to'''
        return self._block_class_client._branch_client

    @property
    def _block_openapi_instance(self) -> Api:
        '''The api instance instantiated with the appropriate `SedaroApiClient` to interact with when CRUDing Blocks'''
        return self._block_class_client._block_openapi_instance

    @property
    def _sedaro_client(self) -> 'SedaroApiClient':
        '''The `SedaroApiClient` this `BlockClient` was accessed through'''
        return self._branch_client._sedaro_client

    def check_still_exists(self) -> bool:
        """Checks whether the Sedaro Block this `BlockClient` references still exists.

        Returns:
            bool: indication of whether or not the referenced Sedaro Block still exists
        """
        return self.id in self._branch_client.data[self._block_group]

    def enforce_still_exists(self) -> None:
        """Raises and error if the Sedaro Block this `BlockClient` references no longer exists.

        Raises:
            NonexistantBlockError: indication that the Block no longer exists.
        """
        if not self.check_still_exists():
            raise NonexistantBlockError(
                f'The referenced "{self._block_name}" (id: {self.id}) no longer exists.'
            )

    def update(self, timeout: Union[int, Tuple] = None, **attrs_to_update) -> 'BlockClient':
        """Update attributes of the corresponding Sedaro Block

        Args:
            timeout (Union[int, Tuple], optional): the timeout used by the REST client. Defaults to `None`.
            **attrs_to_update (Dict): desired attributes to update on the Sedaro Block

        Returns:
            BlockClient: updated `BlockClient` (Note: the previous `BlockClient` reference is also updated)
        """
        # NOTE: `self.data` calls `self.enforce_still_exists()`, so don't need to call here
        body = {**self.data, **attrs_to_update}

        # res = getattr(self._block_openapi_instance, f'{UPDATE}_{snake_case(self._block_name)}')( # TODO: temp_crud
        res = temp_crud(self._sedaro_client, 'PATCH')(getattr(self._block_openapi_instance, f'{UPDATE}_{snake_case(self._block_name)}'))(
            # body=self._block_class_client._update_class(**body), # TODO: temp_crud
            body=body,
            path_params={
                'branchId': self._branch_client.id,
                'blockId': int(self.id)
            },
            timeout=timeout
        )
        self._branch_client._process_block_crud_response(res)
        return self

    def delete(self) -> str:
        """Deletes the associated Sedaro Block

        Returns:
            str: `id` of the deleted Sedaro Block
        """
        self.enforce_still_exists()

        id = self.id
        # res = getattr(self._block_openapi_instance, f'{DELETE}_{snake_case(self._block_name)}')( # TODO: temp_crud
        res = temp_crud(self._sedaro_client, 'DELETE')(getattr(self._block_openapi_instance, f'{DELETE}_{snake_case(self._block_name)}'))(
            path_params={
                'branchId': self._branch_client.id,
                "blockId": int(id)
            }
        )
        return self._branch_client._process_block_crud_response(res)

    def get_schema(self) -> dict:
        """Gets the schema for the associated Sedaro Block type

        Returns:
            dict: the associated Block's schema
        """
        return self._branch_client.data_schema['definitions'][self._block_name]

    @lru_cache(maxsize=None)
    def get_field_schema(self, field: str) -> dict:
        """Gets the field schema of the corresponding attribute on the Sedaro Block.

        Args:
            field (str): field to get the schema for

        Raises:
            TypeError: if the value of `field` is not a string
            AttributeError: if the value of `field` does not correspond to any field on the associated Sedaro Block

        Returns:
            dict: the field schema
        """
        if type(field) is not str:
            raise TypeError(
                f'The value of the "{field}" argument must be a string, not a {type(field).__name__}')

        properties = self.get_schema()['properties']

        if field not in properties:
            raise make_attr_error(field, self._block_name)

        return properties[field]

    @lru_cache(maxsize=None)
    def is_rel_field(self, field: str) -> bool:
        """Checks if the given `field` is a relationship field on the associated Sedaro Block.

        Args:
            field (str): field to check

        Raises:
            TypeError: if the value of `field` is not a string
            KeyError: if the value of `field` does not correspond to any field on the associated Sedaro Block

        Returns:
            bool: indicates if the given `field` is a relationship field on the Sedaro Block or not.
        """
        field_schema = self.get_field_schema(field)

        title: str = field_schema.get('title')
        description: str = field_schema.get('description')

        # all rel fields have a title and description
        if None in (title, description):
            return False

        # all rel fields have these strings in title and description
        return 'ID' in title and all(s in description for s in ['Relationship', '`', 'block', 'On delete'])

    def get_rel_field_type(self, field: str) -> str:
        """Get the type of relationship of the field. Note: first call `is_rel_field` if you need to confirm `field` is
        a relationship field.

        Args:
            field (str): the field to get the relationship type for

        Raises:
            TypeError: if the value of `field` is not a string or not a relationship field on this type of Sedaro Block
            KeyError: if the value of `field` does not correspond to any field on the associated Sedaro Block
            NotImplementedError: if the relationship type is not able to be parsed from the field schema

        Returns:
            str: a string indicating the type of relationship field
        """
        if not self.is_rel_field(field):
            raise TypeError(
                f'The given field "{field}" is not a relationship field on "{self._block_name}".')

        description: str = self.get_field_schema(field).get('description')

        # NOTE: order of if statements matters
        if 'with data to one or more' in description:
            return MANY_SIDE_DATA

        if 'to one or more' in description:
            return MANY_SIDE

        if any(s in description for s in ['to a', 'to zero or one']):
            return ONE_SIDE

        raise NotImplementedError(
            f'Unsupported relationship type on "{self._block_name}", attribute: "{field}".'
        )


# ------ helper function and vars for this file only ------
MANY_SIDE = 'many-side'
MANY_SIDE_DATA = 'many-side-with-data'
ONE_SIDE = 'one-side'


def make_attr_error(field: str, block_name: str) -> str:
    return AttributeError(f'There is no "{field}" attribute on "{block_name}"')



from sedaro_base_client.paths.models_branches_branch_id_system_geometry_body_frame_vectors_.post import ApiForpost


class ModelsBranchesBranchIdSystemGeometryBodyFrameVectors(
    ApiForpost,
):
    pass



from sedaro_base_client.paths.models_branches_branch_id_gnc_algorithms_orbit_determination_ekf_.post import ApiForpost


class ModelsBranchesBranchIdGncAlgorithmsOrbitDeterminationEkf(
    ApiForpost,
):
    pass



from sedaro_base_client.paths.models_branches_branch_id_cdh_conops_ground_targets_block_id.delete import ApiFordelete
from sedaro_base_client.paths.models_branches_branch_id_cdh_conops_ground_targets_block_id.patch import ApiForpatch


class ModelsBranchesBranchIdCdhConopsGroundTargetsBlockId(
    ApiFordelete,
    ApiForpatch,
):
    pass



from sedaro_base_client.paths.models_branches_branch_id_gnc_sensors_fields_of_view_constraints_.post import ApiForpost


class ModelsBranchesBranchIdGncSensorsFieldsOfViewConstraints(
    ApiForpost,
):
    pass




import gzip
import json
import time

import datetime as dt

from typing import List, Union
from pathlib import Path
from sedaro import SedaroApiClient
from sedaro_base_client.apis.tags import jobs_api
from sedaro.results.utils import (
    _get_agent_id_name_map, _restructure_data, progress_bar, hfill, HFILL, STATUS_ICON_MAP
)
from sedaro.results.agent import SedaroAgentResult


class SedaroSimulationResult:

    def __init__(self, simulation: dict, data: dict):
        '''Initialize a new Simulation Result.

        See the following class methods for simple initialization:
            - get_scenario_latest
            - poll_scenario_latest
            - from_file
        '''
        self.__simulation = {
            'branch': int(simulation['branch']),
            'simulatedAgents': dict(simulation['simulatedAgents']),
            'dateCreated': simulation['dateCreated'],
            'dateModified': simulation['dateModified'],
            'status': str(simulation['status']),
        }
        self.__data = data
        self.__branch = simulation['branch']
        if self.success:
            self.__meta = data['Data']['meta']
            raw_series = data['Data']['series']
            agent_id_name_map = _get_agent_id_name_map(self.__meta)
            agent_simbed_id_map = {value: key for key, value in self.__simulation['simulatedAgents'].items()}
            self.__simpleseries, self._agent_blocks = _restructure_data(
                raw_series, agent_id_name_map, self.__meta, agent_simbed_id_map
            )
        else:
            self.__meta = None
            self.__simpleseries = None
            self._agent_blocks = None

    def __repr__(self) -> str:
        return f'SedaroSimulationResult(branch={self.__branch}, status={self.status})'

    @classmethod
    def get_scenario_latest(cls, api_key: str, scenario_id: int, host: str = 'https://api.sedaro.com'):
        '''Query latest scenario result.'''
        with SedaroApiClient(api_key=api_key, host=host) as sedaro_client:
            api_instance = jobs_api.JobsApi(sedaro_client)
            simulation = cls.__get_simulation(api_instance, scenario_id)
            if simulation['status'] == 'SUCCEEDED':
                data = sedaro_client.get_data(simulation['dataId'])
            else:
                data = None
            return cls(simulation, data)

    @classmethod
    def poll_scenario_latest(
        cls,
        api_key: str,
        scenario_id: int,
        host: str = 'https://api.sedaro.com',
        retry_interval: int = 2
    ):
        '''Query latest scenario result and wait for sim if it is running.'''
        with SedaroApiClient(api_key=api_key, host=host) as sedaro_client:
            api_instance = jobs_api.JobsApi(sedaro_client)
            simulation = cls.__get_simulation(api_instance, scenario_id)

            while simulation['status'] in ('PENDING', 'RUNNING'):
                simulation = cls.__get_simulation(api_instance, scenario_id)
                progress_bar(simulation['progress']['percentComplete'])
                time.sleep(retry_interval)

            return cls.get_scenario_latest(api_key, scenario_id, host=host)

    @staticmethod
    def __get_simulation(api_instance, scenario_id: int) -> dict:
        try:
            return api_instance.get_simulations(
                path_params={'branchId': scenario_id},
                query_params={'latest': ''}
            ).body[0]
        except IndexError:
            raise IndexError(f'Could not find any simulation results for scenario: {scenario_id}')

    @property
    def templated_agents(self) -> List[str]:
        return tuple([
            entry['name'] for _, entry
            in self.__meta['structure']['scenario']['Agent'].items()
            if not entry['peripheral']
        ])

    @property
    def peripheral_agents(self) -> List[str]:
        return tuple([
            entry['name'] for _, entry
            in self.__meta['structure']['scenario']['Agent'].items()
            if entry['peripheral']
        ])

    @property
    def status(self) -> str:
        return str(self.__simulation['status'])

    @property
    def start_time(self) -> dt.datetime:
        return dt.datetime.strptime(self.__simulation['dateCreated'], "%Y-%m-%dT%H:%M:%S.%fZ")

    @property
    def end_time(self) -> dt.datetime:
        return dt.datetime.strptime(self.__simulation['dateModified'], "%Y-%m-%dT%H:%M:%S.%fZ")

    @property
    def run_time(self) -> float:
        return (self.end_time - self.start_time).total_seconds()

    @property
    def success(self) -> bool:
        return str(self.__simulation['status']) == 'SUCCEEDED'

    def __assert_success(self) -> None:
        if not self.success:
            raise ValueError('This operation cannot be completed because the simulation failed.')

    def __agent_id_from_name(self, name: str) -> str:
        for id_, entry in self.__meta['structure']['scenario']['Agent'].items():
            if name == entry['name']:
                return id_
        else:
            raise ValueError(f"Agent {name} not found in data set.")

    def agent(self, name: str) -> SedaroAgentResult:
        '''Query results for a particular agent by name.'''
        self.__assert_success()
        agent_id = self.__agent_id_from_name(name)
        return SedaroAgentResult(name, self._agent_blocks[agent_id], self.__simpleseries[name])

    def to_file(self, filename: Union[str, Path]) -> None:
        '''Save simulation result to compressed JSON file.'''
        with gzip.open(filename, 'xt', encoding='UTF-8') as json_file:
            contents = {'data': self.__data, 'simulation': self.__simulation}
            json.dump(contents, json_file)
            print(f"💾 Successfully saved to {filename}")

    @classmethod
    def from_file(cls, filename: Union[str, Path]):
        '''Load simulation result from compressed JSON file.'''
        with gzip.open(filename, 'rt', encoding='UTF-8') as json_file:
            contents = json.load(json_file)
            return cls(contents['simulation'], contents['data'])

    def summarize(self) -> None:
        '''Summarize these results in the console.'''
        hfill()
        print(f'Sedaro Simulation Result Summary'.center(HFILL))
        hfill()
        print(f'{STATUS_ICON_MAP[self.status]} Simulation {self.status.lower()} after {self.run_time:.1f}s')

        agents = self.templated_agents
        if len(agents) > 0:
            print('\n🛰️ Templated Agents ')
            for entry in agents:
                print(f'    • {entry}')

        agents = self.peripheral_agents
        if len(agents) > 0:
            print('\n📡 Peripheral Agents ')
            for entry in agents:
                print(f'    • {entry}')

        hfill()
        print("❓ Query agent results with .agent(<NAME>)")



# coding: utf-8

# flake8: noqa

# import all models into this package
# if you have many models here with many references from one model to another this may
# raise a RecursionError
# to avoid this, import only the models that you directly need like:
# from sedaro_base_client.model.pet import Pet
# or import this package, but before doing it, use:
# import sys
# sys.setrecursionlimit(n)

from sedaro_base_client.model.branch_create import BranchCreate
from sedaro_base_client.model.branch_delete_res import BranchDeleteRes
from sedaro_base_client.model.branch_merge import BranchMerge
from sedaro_base_client.model.branch_merge_conflicts_res import BranchMergeConflictsRes
from sedaro_base_client.model.branch_scenario_res import BranchScenarioRes
from sedaro_base_client.model.branch_update import BranchUpdate
from sedaro_base_client.model.branch_vehicle_res import BranchVehicleRes
from sedaro_base_client.model.branch_verify_password import BranchVerifyPassword
from sedaro_base_client.model.conflicts_obj import ConflictsObj
from sedaro_base_client.model.data_service_response import DataServiceResponse
from sedaro_base_client.model.data_set import DataSet
from sedaro_base_client.model.deleted_entity import DeletedEntity
from sedaro_base_client.model.http_validation_error import HTTPValidationError
from sedaro_base_client.model.message_res import MessageRes
from sedaro_base_client.model.simulation_job import SimulationJob
from sedaro_base_client.model.statuses import Statuses
from sedaro_base_client.model.validation_error import ValidationError



from sedaro_base_client.paths.models_branches_branch_id_gnc_fuel_reservoirs_tanks_spherical_.post import ApiForpost


class ModelsBranchesBranchIdGncFuelReservoirsTanksSpherical(
    ApiForpost,
):
    pass



from sedaro_base_client.paths.models_branches_branch_id_gnc_sensors_angular_velocity_sensors_.post import ApiForpost


class ModelsBranchesBranchIdGncSensorsAngularVelocitySensors(
    ApiForpost,
):
    pass



# coding: utf-8

"""
    Sedaro Satellite API

     Allows for consumption of Sedaro Satellite services. Read more about Sedaro Satellite at [docs.sedaro.com](https://docs.sedaro.com).  ### Clients  **Python:** [sedaro](https://pypi.org/project/sedaro/) - This package provides additional functionality on top of the auto-generated OpenAPI client. See the package docs for more information.  ### API Key  To access the Sedaro service via this API, you will need an API key.  You can generate an API key for your account in the  Sedaro [Management Console](https://satellite.sedaro.com/#/account). Once complete, pass the API key in all requests  via the `X_API_KEY` HTTP header.  *API keys grant full access to your account and should never be shared. If you think your API key has been compromised,  you can revoke it in the [Management Console](https://satellite.sedaro.com/#/account).*  ### Jupyter Notebooks  For additional examples of how to use this API for modeling and simulation, see our [Mod-Sim Notebooks](https://github.com/sedaro/modsim-notebooks).  ### Community, Support, Discussion  If you have any issues or suggestions, please reach out:  1. Join the Sedaro Community [Slack](https://join.slack.com/t/sedaro-community/shared_invite/zt-1jps4i711-mXy88AZQ9AV7YcEXr8x7Ow) 2. Email us at support@sedarotech.com  ### Known Issues  - Currently the documentation for 200 responses to Block create, read, update, and delete (CRUD) operations is incorrect. This is due to an issue with our documentation generator.  Under each Block Group, the documentation will show `name`, `collection`, and `data` keys.  In reality, this level does not exist and should be skipped.  See the schema under the `data` key of a Template's Block Group for the correct schema of such Block Group. - Error responses are more specific than what is shown throughout the documentation.  A 4xx or 5xx error will be returned in all error cases.  Only a `200` status indicates success.  See a given error response for additional details.   # noqa: E501

    The version of the OpenAPI document: 3.5.7
    Generated by: https://openapi-generator.tech
"""
import dataclasses
import typing

from urllib3._collections import HTTPHeaderDict


class OpenApiException(Exception):
    """The base exception class for all OpenAPIExceptions"""


class ApiTypeError(OpenApiException, TypeError):
    def __init__(self, msg, path_to_item=None, valid_classes=None,
                 key_type=None):
        """ Raises an exception for TypeErrors

        Args:
            msg (str): the exception message

        Keyword Args:
            path_to_item (list): a list of keys an indices to get to the
                                 current_item
                                 None if unset
            valid_classes (tuple): the primitive classes that current item
                                   should be an instance of
                                   None if unset
            key_type (bool): False if our value is a value in a dict
                             True if it is a key in a dict
                             False if our item is an item in a list
                             None if unset
        """
        self.path_to_item = path_to_item
        self.valid_classes = valid_classes
        self.key_type = key_type
        full_msg = msg
        if path_to_item:
            full_msg = "{0} at {1}".format(msg, render_path(path_to_item))
        super(ApiTypeError, self).__init__(full_msg)


class ApiValueError(OpenApiException, ValueError):
    def __init__(self, msg, path_to_item=None):
        """
        Args:
            msg (str): the exception message

        Keyword Args:
            path_to_item (list) the path to the exception in the
                received_data dict. None if unset
        """

        self.path_to_item = path_to_item
        full_msg = msg
        if path_to_item:
            full_msg = "{0} at {1}".format(msg, render_path(path_to_item))
        super(ApiValueError, self).__init__(full_msg)


class ApiAttributeError(OpenApiException, AttributeError):
    def __init__(self, msg, path_to_item=None):
        """
        Raised when an attribute reference or assignment fails.

        Args:
            msg (str): the exception message

        Keyword Args:
            path_to_item (None/list) the path to the exception in the
                received_data dict
        """
        self.path_to_item = path_to_item
        full_msg = msg
        if path_to_item:
            full_msg = "{0} at {1}".format(msg, render_path(path_to_item))
        super(ApiAttributeError, self).__init__(full_msg)


class ApiKeyError(OpenApiException, KeyError):
    def __init__(self, msg, path_to_item=None):
        """
        Args:
            msg (str): the exception message

        Keyword Args:
            path_to_item (None/list) the path to the exception in the
                received_data dict
        """
        self.path_to_item = path_to_item
        full_msg = msg
        if path_to_item:
            full_msg = "{0} at {1}".format(msg, render_path(path_to_item))
        super(ApiKeyError, self).__init__(full_msg)


T = typing.TypeVar("T")


@dataclasses.dataclass
class ApiException(OpenApiException, typing.Generic[T]):
    status: int
    reason: str
    api_response: typing.Optional[T] = None

    @property
    def body(self) -> typing.Union[str, bytes, None]:
        if not self.api_response:
            return None
        return self.api_response.response.data

    @property
    def headers(self) -> typing.Optional[HTTPHeaderDict]:
        if not self.api_response:
            return None
        return self.api_response.response.getheaders()

    def __str__(self):
        """Custom error messages for exception"""
        error_message = "({0})\n"\
                        "Reason: {1}\n".format(self.status, self.reason)
        if self.headers:
            error_message += "HTTP response headers: {0}\n".format(
                self.headers)

        if self.body:
            error_message += "HTTP response body: {0}\n".format(self.body)

        return error_message


def render_path(path_to_item):
    """Returns a string representation of a path"""
    result = ""
    for pth in path_to_item:
        if isinstance(pth, int):
            result += "[{0}]".format(pth)
        else:
            result += "['{0}']".format(pth)
    return result



# do not import all endpoints into this module because that uses a lot of memory and stack frames
# if you need the ability to import all endpoints from this module, import them with
# from sedaro_base_client.paths.models_branches_branch_id_thermal_thermal_interfaces_block_id import Api

from sedaro_base_client.paths import PathValues

path = PathValues.MODELS_BRANCHES_BRANCH_ID_THERMAL_THERMALINTERFACES_BLOCK_ID


import typing_extensions

from sedaro_base_client.apis.tags import TagValues
from sedaro_base_client.apis.tags.branches_api import BranchesApi
from sedaro_base_client.apis.tags.data_api import DataApi
from sedaro_base_client.apis.tags.agent_api import AgentApi
from sedaro_base_client.apis.tags.angular_velocity_sensor_api import AngularVelocitySensorApi
from sedaro_base_client.apis.tags.antenna_api import AntennaApi
from sedaro_base_client.apis.tags.averaging_algorithm_api import AveragingAlgorithmApi
from sedaro_base_client.apis.tags.battery_api import BatteryApi
from sedaro_base_client.apis.tags.battery_cell_api import BatteryCellApi
from sedaro_base_client.apis.tags.battery_pack_api import BatteryPackApi
from sedaro_base_client.apis.tags.body_frame_vector_api import BodyFrameVectorApi
from sedaro_base_client.apis.tags.bus_regulator_api import BusRegulatorApi
from sedaro_base_client.apis.tags.celestial_target_api import CelestialTargetApi
from sedaro_base_client.apis.tags.celestial_vector_api import CelestialVectorApi
from sedaro_base_client.apis.tags.circular_field_of_view_api import CircularFieldOfViewApi
from sedaro_base_client.apis.tags.clock_config_api import ClockConfigApi
from sedaro_base_client.apis.tags.component_api import ComponentApi
from sedaro_base_client.apis.tags.condition_api import ConditionApi
from sedaro_base_client.apis.tags.constant_load_api import ConstantLoadApi
from sedaro_base_client.apis.tags.cooler_api import CoolerApi
from sedaro_base_client.apis.tags.cooperative_transmit_interface_api import CooperativeTransmitInterfaceApi
from sedaro_base_client.apis.tags.data_bus_api import DataBusApi
from sedaro_base_client.apis.tags.data_mode_api import DataModeApi
from sedaro_base_client.apis.tags.data_storage_api import DataStorageApi
from sedaro_base_client.apis.tags.data_type_api import DataTypeApi
from sedaro_base_client.apis.tags.direction_sensor_api import DirectionSensorApi
from sedaro_base_client.apis.tags.ekf_algorithm_api import EKFAlgorithmApi
from sedaro_base_client.apis.tags.fov_constraint_api import FOVConstraintApi
from sedaro_base_client.apis.tags.fuel_reservoir_api import FuelReservoirApi
from sedaro_base_client.apis.tags.gps_algorithm_api import GPSAlgorithmApi
from sedaro_base_client.apis.tags.ground_target_api import GroundTargetApi
from sedaro_base_client.apis.tags.group_condition_api import GroupConditionApi
from sedaro_base_client.apis.tags.heater_api import HeaterApi
from sedaro_base_client.apis.tags.internal_data_interface_api import InternalDataInterfaceApi
from sedaro_base_client.apis.tags.jobs_api import JobsApi
from sedaro_base_client.apis.tags.laser_comm_module_api import LaserCommModuleApi
from sedaro_base_client.apis.tags.load_state_api import LoadStateApi
from sedaro_base_client.apis.tags.local_vector_api import LocalVectorApi
from sedaro_base_client.apis.tags.lock_pointing_mode_api import LockPointingModeApi
from sedaro_base_client.apis.tags.mekf_algorithm_api import MEKFAlgorithmApi
from sedaro_base_client.apis.tags.magnetorquer_api import MagnetorquerApi
from sedaro_base_client.apis.tags.max_align_pointing_mode_api import MaxAlignPointingModeApi
from sedaro_base_client.apis.tags.modem_api import ModemApi
from sedaro_base_client.apis.tags.operational_mode_api import OperationalModeApi
from sedaro_base_client.apis.tags.optical_attitude_sensor_api import OpticalAttitudeSensorApi
from sedaro_base_client.apis.tags.orbit_api import OrbitApi
from sedaro_base_client.apis.tags.pid_algorithm_api import PIDAlgorithmApi
from sedaro_base_client.apis.tags.passive_pointing_mode_api import PassivePointingModeApi
from sedaro_base_client.apis.tags.passive_transmit_interface_api import PassiveTransmitInterfaceApi
from sedaro_base_client.apis.tags.position_sensor_api import PositionSensorApi
from sedaro_base_client.apis.tags.power_processor_api import PowerProcessorApi
from sedaro_base_client.apis.tags.reaction_wheel_api import ReactionWheelApi
from sedaro_base_client.apis.tags.receive_interface_api import ReceiveInterfaceApi
from sedaro_base_client.apis.tags.rectangular_field_of_view_api import RectangularFieldOfViewApi
from sedaro_base_client.apis.tags.satellite_api import SatelliteApi
from sedaro_base_client.apis.tags.sliding_mode_algorithm_api import SlidingModeAlgorithmApi
from sedaro_base_client.apis.tags.solar_array_api import SolarArrayApi
from sedaro_base_client.apis.tags.solar_cell_api import SolarCellApi
from sedaro_base_client.apis.tags.solar_panel_api import SolarPanelApi
from sedaro_base_client.apis.tags.space_target_api import SpaceTargetApi
from sedaro_base_client.apis.tags.spherical_fuel_tank_api import SphericalFuelTankApi
from sedaro_base_client.apis.tags.spherocylinder_fuel_tank_api import SpherocylinderFuelTankApi
from sedaro_base_client.apis.tags.static_thrust_control_algorithm_api import StaticThrustControlAlgorithmApi
from sedaro_base_client.apis.tags.subsystem_api import SubsystemApi
from sedaro_base_client.apis.tags.surface_api import SurfaceApi
from sedaro_base_client.apis.tags.surface_material_api import SurfaceMaterialApi
from sedaro_base_client.apis.tags.target_group_api import TargetGroupApi
from sedaro_base_client.apis.tags.target_group_vector_api import TargetGroupVectorApi
from sedaro_base_client.apis.tags.target_vector_api import TargetVectorApi
from sedaro_base_client.apis.tags.temp_controller_state_api import TempControllerStateApi
from sedaro_base_client.apis.tags.thermal_interface_api import ThermalInterfaceApi
from sedaro_base_client.apis.tags.thermal_interface_material_api import ThermalInterfaceMaterialApi
from sedaro_base_client.apis.tags.thruster_api import ThrusterApi
from sedaro_base_client.apis.tags.triad_algorithm_api import TriadAlgorithmApi
from sedaro_base_client.apis.tags.vector_sensor_api import VectorSensorApi

TagToApi = typing_extensions.TypedDict(
    'TagToApi',
    {
        TagValues.BRANCHES: BranchesApi,
        TagValues.DATA: DataApi,
        TagValues.AGENT: AgentApi,
        TagValues.ANGULAR_VELOCITY_SENSOR: AngularVelocitySensorApi,
        TagValues.ANTENNA: AntennaApi,
        TagValues.AVERAGING_ALGORITHM: AveragingAlgorithmApi,
        TagValues.BATTERY: BatteryApi,
        TagValues.BATTERY_CELL: BatteryCellApi,
        TagValues.BATTERY_PACK: BatteryPackApi,
        TagValues.BODY_FRAME_VECTOR: BodyFrameVectorApi,
        TagValues.BUS_REGULATOR: BusRegulatorApi,
        TagValues.CELESTIAL_TARGET: CelestialTargetApi,
        TagValues.CELESTIAL_VECTOR: CelestialVectorApi,
        TagValues.CIRCULAR_FIELD_OF_VIEW: CircularFieldOfViewApi,
        TagValues.CLOCK_CONFIG: ClockConfigApi,
        TagValues.COMPONENT: ComponentApi,
        TagValues.CONDITION: ConditionApi,
        TagValues.CONSTANT_LOAD: ConstantLoadApi,
        TagValues.COOLER: CoolerApi,
        TagValues.COOPERATIVE_TRANSMIT_INTERFACE: CooperativeTransmitInterfaceApi,
        TagValues.DATA_BUS: DataBusApi,
        TagValues.DATA_MODE: DataModeApi,
        TagValues.DATA_STORAGE: DataStorageApi,
        TagValues.DATA_TYPE: DataTypeApi,
        TagValues.DIRECTION_SENSOR: DirectionSensorApi,
        TagValues.EKF_ALGORITHM: EKFAlgorithmApi,
        TagValues.FOV_CONSTRAINT: FOVConstraintApi,
        TagValues.FUEL_RESERVOIR: FuelReservoirApi,
        TagValues.GPS_ALGORITHM: GPSAlgorithmApi,
        TagValues.GROUND_TARGET: GroundTargetApi,
        TagValues.GROUP_CONDITION: GroupConditionApi,
        TagValues.HEATER: HeaterApi,
        TagValues.INTERNAL_DATA_INTERFACE: InternalDataInterfaceApi,
        TagValues.JOBS: JobsApi,
        TagValues.LASER_COMM_MODULE: LaserCommModuleApi,
        TagValues.LOAD_STATE: LoadStateApi,
        TagValues.LOCAL_VECTOR: LocalVectorApi,
        TagValues.LOCK_POINTING_MODE: LockPointingModeApi,
        TagValues.MEKF_ALGORITHM: MEKFAlgorithmApi,
        TagValues.MAGNETORQUER: MagnetorquerApi,
        TagValues.MAX_ALIGN_POINTING_MODE: MaxAlignPointingModeApi,
        TagValues.MODEM: ModemApi,
        TagValues.OPERATIONAL_MODE: OperationalModeApi,
        TagValues.OPTICAL_ATTITUDE_SENSOR: OpticalAttitudeSensorApi,
        TagValues.ORBIT: OrbitApi,
        TagValues.PID_ALGORITHM: PIDAlgorithmApi,
        TagValues.PASSIVE_POINTING_MODE: PassivePointingModeApi,
        TagValues.PASSIVE_TRANSMIT_INTERFACE: PassiveTransmitInterfaceApi,
        TagValues.POSITION_SENSOR: PositionSensorApi,
        TagValues.POWER_PROCESSOR: PowerProcessorApi,
        TagValues.REACTION_WHEEL: ReactionWheelApi,
        TagValues.RECEIVE_INTERFACE: ReceiveInterfaceApi,
        TagValues.RECTANGULAR_FIELD_OF_VIEW: RectangularFieldOfViewApi,
        TagValues.SATELLITE: SatelliteApi,
        TagValues.SLIDING_MODE_ALGORITHM: SlidingModeAlgorithmApi,
        TagValues.SOLAR_ARRAY: SolarArrayApi,
        TagValues.SOLAR_CELL: SolarCellApi,
        TagValues.SOLAR_PANEL: SolarPanelApi,
        TagValues.SPACE_TARGET: SpaceTargetApi,
        TagValues.SPHERICAL_FUEL_TANK: SphericalFuelTankApi,
        TagValues.SPHEROCYLINDER_FUEL_TANK: SpherocylinderFuelTankApi,
        TagValues.STATIC_THRUST_CONTROL_ALGORITHM: StaticThrustControlAlgorithmApi,
        TagValues.SUBSYSTEM: SubsystemApi,
        TagValues.SURFACE: SurfaceApi,
        TagValues.SURFACE_MATERIAL: SurfaceMaterialApi,
        TagValues.TARGET_GROUP: TargetGroupApi,
        TagValues.TARGET_GROUP_VECTOR: TargetGroupVectorApi,
        TagValues.TARGET_VECTOR: TargetVectorApi,
        TagValues.TEMP_CONTROLLER_STATE: TempControllerStateApi,
        TagValues.THERMAL_INTERFACE: ThermalInterfaceApi,
        TagValues.THERMAL_INTERFACE_MATERIAL: ThermalInterfaceMaterialApi,
        TagValues.THRUSTER: ThrusterApi,
        TagValues.TRIAD_ALGORITHM: TriadAlgorithmApi,
        TagValues.VECTOR_SENSOR: VectorSensorApi,
    }
)

tag_to_api = TagToApi(
    {
        TagValues.BRANCHES: BranchesApi,
        TagValues.DATA: DataApi,
        TagValues.AGENT: AgentApi,
        TagValues.ANGULAR_VELOCITY_SENSOR: AngularVelocitySensorApi,
        TagValues.ANTENNA: AntennaApi,
        TagValues.AVERAGING_ALGORITHM: AveragingAlgorithmApi,
        TagValues.BATTERY: BatteryApi,
        TagValues.BATTERY_CELL: BatteryCellApi,
        TagValues.BATTERY_PACK: BatteryPackApi,
        TagValues.BODY_FRAME_VECTOR: BodyFrameVectorApi,
        TagValues.BUS_REGULATOR: BusRegulatorApi,
        TagValues.CELESTIAL_TARGET: CelestialTargetApi,
        TagValues.CELESTIAL_VECTOR: CelestialVectorApi,
        TagValues.CIRCULAR_FIELD_OF_VIEW: CircularFieldOfViewApi,
        TagValues.CLOCK_CONFIG: ClockConfigApi,
        TagValues.COMPONENT: ComponentApi,
        TagValues.CONDITION: ConditionApi,
        TagValues.CONSTANT_LOAD: ConstantLoadApi,
        TagValues.COOLER: CoolerApi,
        TagValues.COOPERATIVE_TRANSMIT_INTERFACE: CooperativeTransmitInterfaceApi,
        TagValues.DATA_BUS: DataBusApi,
        TagValues.DATA_MODE: DataModeApi,
        TagValues.DATA_STORAGE: DataStorageApi,
        TagValues.DATA_TYPE: DataTypeApi,
        TagValues.DIRECTION_SENSOR: DirectionSensorApi,
        TagValues.EKF_ALGORITHM: EKFAlgorithmApi,
        TagValues.FOV_CONSTRAINT: FOVConstraintApi,
        TagValues.FUEL_RESERVOIR: FuelReservoirApi,
        TagValues.GPS_ALGORITHM: GPSAlgorithmApi,
        TagValues.GROUND_TARGET: GroundTargetApi,
        TagValues.GROUP_CONDITION: GroupConditionApi,
        TagValues.HEATER: HeaterApi,
        TagValues.INTERNAL_DATA_INTERFACE: InternalDataInterfaceApi,
        TagValues.JOBS: JobsApi,
        TagValues.LASER_COMM_MODULE: LaserCommModuleApi,
        TagValues.LOAD_STATE: LoadStateApi,
        TagValues.LOCAL_VECTOR: LocalVectorApi,
        TagValues.LOCK_POINTING_MODE: LockPointingModeApi,
        TagValues.MEKF_ALGORITHM: MEKFAlgorithmApi,
        TagValues.MAGNETORQUER: MagnetorquerApi,
        TagValues.MAX_ALIGN_POINTING_MODE: MaxAlignPointingModeApi,
        TagValues.MODEM: ModemApi,
        TagValues.OPERATIONAL_MODE: OperationalModeApi,
        TagValues.OPTICAL_ATTITUDE_SENSOR: OpticalAttitudeSensorApi,
        TagValues.ORBIT: OrbitApi,
        TagValues.PID_ALGORITHM: PIDAlgorithmApi,
        TagValues.PASSIVE_POINTING_MODE: PassivePointingModeApi,
        TagValues.PASSIVE_TRANSMIT_INTERFACE: PassiveTransmitInterfaceApi,
        TagValues.POSITION_SENSOR: PositionSensorApi,
        TagValues.POWER_PROCESSOR: PowerProcessorApi,
        TagValues.REACTION_WHEEL: ReactionWheelApi,
        TagValues.RECEIVE_INTERFACE: ReceiveInterfaceApi,
        TagValues.RECTANGULAR_FIELD_OF_VIEW: RectangularFieldOfViewApi,
        TagValues.SATELLITE: SatelliteApi,
        TagValues.SLIDING_MODE_ALGORITHM: SlidingModeAlgorithmApi,
        TagValues.SOLAR_ARRAY: SolarArrayApi,
        TagValues.SOLAR_CELL: SolarCellApi,
        TagValues.SOLAR_PANEL: SolarPanelApi,
        TagValues.SPACE_TARGET: SpaceTargetApi,
        TagValues.SPHERICAL_FUEL_TANK: SphericalFuelTankApi,
        TagValues.SPHEROCYLINDER_FUEL_TANK: SpherocylinderFuelTankApi,
        TagValues.STATIC_THRUST_CONTROL_ALGORITHM: StaticThrustControlAlgorithmApi,
        TagValues.SUBSYSTEM: SubsystemApi,
        TagValues.SURFACE: SurfaceApi,
        TagValues.SURFACE_MATERIAL: SurfaceMaterialApi,
        TagValues.TARGET_GROUP: TargetGroupApi,
        TagValues.TARGET_GROUP_VECTOR: TargetGroupVectorApi,
        TagValues.TARGET_VECTOR: TargetVectorApi,
        TagValues.TEMP_CONTROLLER_STATE: TempControllerStateApi,
        TagValues.THERMAL_INTERFACE: ThermalInterfaceApi,
        TagValues.THERMAL_INTERFACE_MATERIAL: ThermalInterfaceMaterialApi,
        TagValues.THRUSTER: ThrusterApi,
        TagValues.TRIAD_ALGORITHM: TriadAlgorithmApi,
        TagValues.VECTOR_SENSOR: VectorSensorApi,
    }
)



import json
from typing import Dict, Optional

from sedaro_base_client import Configuration
from sedaro_base_client.api_client import ApiClient
from sedaro_base_client.apis.tags import branches_api

from .branch_client import BranchClient
from .exceptions import SedaroApiException
from .settings import COMMON_API_KWARGS
from .sim_client import SimClient
from .utils import parse_urllib_response


class SedaroApiClient(ApiClient):
    """A client to interact with the Sedaro API"""

    def __init__(self, api_key, host='https://api.sedaro.com', *args, **kwargs):
        return super().__init__(
            configuration=Configuration(host=host),
            *args,
            **kwargs,
            header_name='X_API_KEY',
            header_value=api_key
        )

    def get_branch(self, id: int) -> BranchClient:
        """Gets a Sedaro Branch based on the given `id` and creates a `BranchClient` from the response. The branch must
        be accessible to this `SedaroApiClient` via the `api_key`.

        Args:
            id (int): the id of the desired Sedaro Branch

        Returns:
            BranchClient: A `BranchClient` object used to interact with the data attached to the corresponding Sedaro
            Branch.
        """
        branches_api_instance = branches_api.BranchesApi(self)
        # res = branches_api_instance.get_branch(path_params={'branchId': id}) # TODO: temp_crud
        # return BranchClient(res.body, self)
        res = branches_api_instance.get_branch(
            path_params={'branchId': id}, **COMMON_API_KWARGS)
        body = res.body
        if COMMON_API_KWARGS['skip_deserialization']:
            body = parse_urllib_response(res.response)
        return BranchClient(body, self)

    def get_data(self, id, start: float = None, stop: float = None, binWidth: float = None, limit: float = None, axisOrder: str = None):
        """Simplified Data Service getter with significantly higher performance over the Swagger-generated client."""
        url = f'/data/?id={id}'
        if start is not None:
            url += f'&start={start}'
        if stop is not None:
            url += f'&stop={stop}'
        if binWidth is not None:
            url += f'&binWidth={binWidth}'
        elif limit is not None:
            url += f'&limit={limit}'
        if axisOrder is not None:
            if axisOrder not in {'TIME_MAJOR',  'TIME_MINOR'}:
                raise ValueError(
                    'axisOrder must be either "TIME_MAJOR" or "TIME_MINOR"')
            url += f'&axisOrder={axisOrder}'
        response = self.call_api(url, 'GET')
        _response = None
        try:
            _response = parse_urllib_response(response)
            if response.status != 200:
                raise Exception()
        except:
            reason = _response['error']['message'] if _response and 'error' in _response else 'An unknown error occurred.'
            raise SedaroApiException(status=response.status, reason=reason)
        return _response

    def get_sim_client(self, branch_id: int):
        """Creates and returns a Sedaro SimClient

        Args:
            branch_id (int): id of the desired Sedaro Scenario Branch to interact with its simulations (jobs)

        Returns:
            SimClient: a Sedaro SimClient
        """
        return SimClient(self, branch_id)

    def send_request(self, resource_path: str, method: str, body: Optional[Dict] = None):
        """Send a request to the Sedaro server

        Args:
            resource_path (str): url path (everything after the host) for desired route
            method (str): HTTP method ('GET', 'POST', 'DELETE'...etc)
            body (Optional[Union[str, bytes]], optional): Body of the request. Defaults to None.

        Returns:
            Dict: dictionary from the response body
        """
        headers = {}
        if body is not None:
            body = json.dumps(body)
            headers['Content-Type'] = 'application/json'
        res = self.call_api(
            resource_path,
            method.upper(),
            headers=headers,
            body=body
        )
        return parse_urllib_response(res)



from sedaro_base_client.paths.models_branches_branch_id_gnc_reference_vectors_celestial_vectors_.post import ApiForpost


class ModelsBranchesBranchIdGncReferenceVectorsCelestialVectors(
    ApiForpost,
):
    pass



# coding: utf-8

"""
    Sedaro Satellite API

     Allows for consumption of Sedaro Satellite services. Read more about Sedaro Satellite at [docs.sedaro.com](https://docs.sedaro.com).  ### Clients  **Python:** [sedaro](https://pypi.org/project/sedaro/) - This package provides additional functionality on top of the auto-generated OpenAPI client. See the package docs for more information.  ### API Key  To access the Sedaro service via this API, you will need an API key.  You can generate an API key for your account in the  Sedaro [Management Console](https://satellite.sedaro.com/#/account). Once complete, pass the API key in all requests  via the `X_API_KEY` HTTP header.  *API keys grant full access to your account and should never be shared. If you think your API key has been compromised,  you can revoke it in the [Management Console](https://satellite.sedaro.com/#/account).*  ### Jupyter Notebooks  For additional examples of how to use this API for modeling and simulation, see our [Mod-Sim Notebooks](https://github.com/sedaro/modsim-notebooks).  ### Community, Support, Discussion  If you have any issues or suggestions, please reach out:  1. Join the Sedaro Community [Slack](https://join.slack.com/t/sedaro-community/shared_invite/zt-1jps4i711-mXy88AZQ9AV7YcEXr8x7Ow) 2. Email us at support@sedarotech.com  ### Known Issues  - Currently the documentation for 200 responses to Block create, read, update, and delete (CRUD) operations is incorrect. This is due to an issue with our documentation generator.  Under each Block Group, the documentation will show `name`, `collection`, and `data` keys.  In reality, this level does not exist and should be skipped.  See the schema under the `data` key of a Template's Block Group for the correct schema of such Block Group. - Error responses are more specific than what is shown throughout the documentation.  A 4xx or 5xx error will be returned in all error cases.  Only a `200` status indicates success.  See a given error response for additional details.   # noqa: E501

    The version of the OpenAPI document: 3.5.7
    Generated by: https://openapi-generator.tech
"""

from sedaro_base_client.paths.models_branches_branch_id_gnc_reference_vectors_celestial_vectors_.post import CreateCelestialVector
from sedaro_base_client.paths.models_branches_branch_id_gnc_reference_vectors_celestial_vectors_block_id.delete import DeleteCelestialVector
from sedaro_base_client.paths.models_branches_branch_id_gnc_reference_vectors_celestial_vectors_block_id.patch import UpdateCelestialVector


class CelestialVectorApi(
    CreateCelestialVector,
    DeleteCelestialVector,
    UpdateCelestialVector,
):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """
    pass



# coding: utf-8

"""
    Sedaro Satellite API

     Allows for consumption of Sedaro Satellite services. Read more about Sedaro Satellite at [docs.sedaro.com](https://docs.sedaro.com).  ### Clients  **Python:** [sedaro](https://pypi.org/project/sedaro/) - This package provides additional functionality on top of the auto-generated OpenAPI client. See the package docs for more information.  ### API Key  To access the Sedaro service via this API, you will need an API key.  You can generate an API key for your account in the  Sedaro [Management Console](https://satellite.sedaro.com/#/account). Once complete, pass the API key in all requests  via the `X_API_KEY` HTTP header.  *API keys grant full access to your account and should never be shared. If you think your API key has been compromised,  you can revoke it in the [Management Console](https://satellite.sedaro.com/#/account).*  ### Jupyter Notebooks  For additional examples of how to use this API for modeling and simulation, see our [Mod-Sim Notebooks](https://github.com/sedaro/modsim-notebooks).  ### Community, Support, Discussion  If you have any issues or suggestions, please reach out:  1. Join the Sedaro Community [Slack](https://join.slack.com/t/sedaro-community/shared_invite/zt-1jps4i711-mXy88AZQ9AV7YcEXr8x7Ow) 2. Email us at support@sedarotech.com  ### Known Issues  - Currently the documentation for 200 responses to Block create, read, update, and delete (CRUD) operations is incorrect. This is due to an issue with our documentation generator.  Under each Block Group, the documentation will show `name`, `collection`, and `data` keys.  In reality, this level does not exist and should be skipped.  See the schema under the `data` key of a Template's Block Group for the correct schema of such Block Group. - Error responses are more specific than what is shown throughout the documentation.  A 4xx or 5xx error will be returned in all error cases.  Only a `200` status indicates success.  See a given error response for additional details.   # noqa: E501

    The version of the OpenAPI document: 3.5.7
    Generated by: https://openapi-generator.tech
"""

import logging
import ssl
from urllib.parse import urlencode
import typing

import certifi
import urllib3
from urllib3._collections import HTTPHeaderDict

from sedaro_base_client.exceptions import ApiException, ApiValueError


logger = logging.getLogger(__name__)


class RESTClientObject(object):

    def __init__(self, configuration, pools_size=4, maxsize=None):
        # urllib3.PoolManager will pass all kw parameters to connectionpool
        # https://github.com/shazow/urllib3/blob/f9409436f83aeb79fbaf090181cd81b784f1b8ce/urllib3/poolmanager.py#L75  # noqa: E501
        # https://github.com/shazow/urllib3/blob/f9409436f83aeb79fbaf090181cd81b784f1b8ce/urllib3/connectionpool.py#L680  # noqa: E501
        # maxsize is the number of requests to host that are allowed in parallel  # noqa: E501
        # Custom SSL certificates and client certificates: http://urllib3.readthedocs.io/en/latest/advanced-usage.html  # noqa: E501

        # cert_reqs
        if configuration.verify_ssl:
            cert_reqs = ssl.CERT_REQUIRED
        else:
            cert_reqs = ssl.CERT_NONE

        # ca_certs
        if configuration.ssl_ca_cert:
            ca_certs = configuration.ssl_ca_cert
        else:
            # if not set certificate file, use Mozilla's root certificates.
            ca_certs = certifi.where()

        addition_pool_args = {}
        if configuration.assert_hostname is not None:
            addition_pool_args['assert_hostname'] = configuration.assert_hostname  # noqa: E501

        if configuration.retries is not None:
            addition_pool_args['retries'] = configuration.retries

        if configuration.socket_options is not None:
            addition_pool_args['socket_options'] = configuration.socket_options

        if maxsize is None:
            if configuration.connection_pool_maxsize is not None:
                maxsize = configuration.connection_pool_maxsize
            else:
                maxsize = 4

        # https pool manager
        if configuration.proxy:
            self.pool_manager = urllib3.ProxyManager(
                num_pools=pools_size,
                maxsize=maxsize,
                cert_reqs=cert_reqs,
                ca_certs=ca_certs,
                cert_file=configuration.cert_file,
                key_file=configuration.key_file,
                proxy_url=configuration.proxy,
                proxy_headers=configuration.proxy_headers,
                **addition_pool_args
            )
        else:
            self.pool_manager = urllib3.PoolManager(
                num_pools=pools_size,
                maxsize=maxsize,
                cert_reqs=cert_reqs,
                ca_certs=ca_certs,
                cert_file=configuration.cert_file,
                key_file=configuration.key_file,
                **addition_pool_args
            )

    def request(
        self,
        method: str,
        url: str,
        headers: typing.Optional[HTTPHeaderDict] = None,
        fields: typing.Optional[typing.Tuple[typing.Tuple[str, typing.Any], ...]] = None,
        body: typing.Optional[typing.Union[str, bytes]] = None,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
    ) -> urllib3.HTTPResponse:
        """Perform requests.

        :param method: http request method
        :param url: http request url
        :param headers: http request headers
        :param body: request body, for other types
        :param fields: request parameters for
                                `application/x-www-form-urlencoded`
                                or `multipart/form-data`
        :param stream: if True, the urllib3.HTTPResponse object will
                                be returned without reading/decoding response
                                data. Default is False.
        :param timeout: timeout setting for this request. If one
                                number provided, it will be total request
                                timeout. It can also be a pair (tuple) of
                                (connection, read) timeouts.
        """
        method = method.upper()
        assert method in ['GET', 'HEAD', 'DELETE', 'POST', 'PUT',
                          'PATCH', 'OPTIONS']

        if fields and body:
            raise ApiValueError(
                "body parameter cannot be used with fields parameter."
            )

        fields = fields or {}
        headers = headers or {}

        if timeout:
            if isinstance(timeout, (int, float)):  # noqa: E501,F821
                timeout = urllib3.Timeout(total=timeout)
            elif (isinstance(timeout, tuple) and
                  len(timeout) == 2):
                timeout = urllib3.Timeout(connect=timeout[0], read=timeout[1])

        try:
            # For `POST`, `PUT`, `PATCH`, `OPTIONS`, `DELETE`
            if method in ['POST', 'PUT', 'PATCH', 'OPTIONS', 'DELETE']:
                if 'Content-Type' not in headers and body is None:
                    r = self.pool_manager.request(
                        method,
                        url,
                        preload_content=not stream,
                        timeout=timeout,
                        headers=headers
                    )
                elif headers['Content-Type'] == 'application/x-www-form-urlencoded':  # noqa: E501
                    r = self.pool_manager.request(
                        method, url,
                        body=body,
                        fields=fields,
                        encode_multipart=False,
                        preload_content=not stream,
                        timeout=timeout,
                        headers=headers)
                elif headers['Content-Type'] == 'multipart/form-data':
                    # must del headers['Content-Type'], or the correct
                    # Content-Type which generated by urllib3 will be
                    # overwritten.
                    del headers['Content-Type']
                    r = self.pool_manager.request(
                        method, url,
                        fields=fields,
                        encode_multipart=True,
                        preload_content=not stream,
                        timeout=timeout,
                        headers=headers)
                # Pass a `string` parameter directly in the body to support
                # other content types than Json when `body` argument is
                # provided in serialized form
                elif isinstance(body, str) or isinstance(body, bytes):
                    request_body = body
                    r = self.pool_manager.request(
                        method, url,
                        body=request_body,
                        preload_content=not stream,
                        timeout=timeout,
                        headers=headers)
                else:
                    # Cannot generate the request from given parameters
                    msg = """Cannot prepare a request message for provided
                             arguments. Please check that your arguments match
                             declared content type."""
                    raise ApiException(status=0, reason=msg)
            # For `GET`, `HEAD`
            else:
                r = self.pool_manager.request(method, url,
                                              preload_content=not stream,
                                              timeout=timeout,
                                              headers=headers)
        except urllib3.exceptions.SSLError as e:
            msg = "{0}\n{1}".format(type(e).__name__, str(e))
            raise ApiException(status=0, reason=msg)

        if not stream:
            # log response body
            logger.debug("response body: %s", r.data)

        return r

    def GET(self, url, headers=None, stream=False,
            timeout=None, fields=None) -> urllib3.HTTPResponse:
        return self.request("GET", url,
                            headers=headers,
                            stream=stream,
                            timeout=timeout,
                            fields=fields)

    def HEAD(self, url, headers=None, stream=False,
             timeout=None, fields=None) -> urllib3.HTTPResponse:
        return self.request("HEAD", url,
                            headers=headers,
                            stream=stream,
                            timeout=timeout,
                            fields=fields)

    def OPTIONS(self, url, headers=None,
                body=None, stream=False, timeout=None, fields=None) -> urllib3.HTTPResponse:
        return self.request("OPTIONS", url,
                            headers=headers,
                            stream=stream,
                            timeout=timeout,
                            body=body, fields=fields)

    def DELETE(self, url, headers=None, body=None,
               stream=False, timeout=None, fields=None) -> urllib3.HTTPResponse:
        return self.request("DELETE", url,
                            headers=headers,
                            stream=stream,
                            timeout=timeout,
                            body=body, fields=fields)

    def POST(self, url, headers=None,
             body=None, stream=False, timeout=None, fields=None) -> urllib3.HTTPResponse:
        return self.request("POST", url,
                            headers=headers,
                            stream=stream,
                            timeout=timeout,
                            body=body, fields=fields)

    def PUT(self, url, headers=None,
            body=None, stream=False, timeout=None, fields=None) -> urllib3.HTTPResponse:
        return self.request("PUT", url,
                            headers=headers,
                            stream=stream,
                            timeout=timeout,
                            body=body, fields=fields)

    def PATCH(self, url, headers=None,
              body=None, stream=False, timeout=None, fields=None) -> urllib3.HTTPResponse:
        return self.request("PATCH", url,
                            headers=headers,
                            stream=stream,
                            timeout=timeout,
                            body=body, fields=fields)



from sedaro_base_client.paths.models_branches_branch_id_gnc_pointing_modes_max_secondary_align_.post import ApiForpost


class ModelsBranchesBranchIdGncPointingModesMaxSecondaryAlign(
    ApiForpost,
):
    pass



from sedaro_base_client.paths.models_branches_branch_id_gnc_fuel_reservoirs_tanks_spherocylinder_block_id.delete import ApiFordelete
from sedaro_base_client.paths.models_branches_branch_id_gnc_fuel_reservoirs_tanks_spherocylinder_block_id.patch import ApiForpatch


class ModelsBranchesBranchIdGncFuelReservoirsTanksSpherocylinderBlockId(
    ApiFordelete,
    ApiForpatch,
):
    pass



import json
import importlib
import inspect
from functools import lru_cache
from pydash.strings import snake_case, pascal_case
from urllib3.response import HTTPResponse
from typing import Dict, Tuple, Union, TYPE_CHECKING
from types import ModuleType
import sys
from dataclasses import dataclass

from sedaro_base_client.api_client import ApiResponse
from .settings import DELETE
from .exceptions import SedaroApiException

if TYPE_CHECKING:
    from .branch_client import BranchClient


@lru_cache(maxsize=None)
def get_snake_and_pascal_case(s: str):
    '''Returns a tuple of the string in snake and pascal case'''
    return snake_case(s), pascal_case(s, strict=False)


def parse_urllib_response(response: HTTPResponse) -> Dict:
    '''Parses the response from urllib3.response.HTTPResponse into a dictionary'''
    return json.loads(response.data.decode('utf-8'))


def parse_block_crud_response(response: ApiResponse) -> Tuple[str, dict, str, str]:
    """Parses the response of CRUD-ing a Sedaro Block on a `Branch` into a tuple.

    Args:
        response (ApiResponse): the response from CRUD-ing a Sedaro Block

    Returns:
        Dict: a tuple of: `block_id`, `block_data`, `block_group`, `action`, `branch_data`, `block_id_to_type_map`
    """
    body = response.body
    action = body['action']
    block_id = str(body['block']['id'])
    block_group = body['block']['group']
    branch_data = body['branch']['data']

    block_data = dict(branch_data[block_group][block_id]) if action.casefold() != DELETE.casefold() \
        else None

    return block_id, block_data, block_group, action, branch_data, body['branch']['blockIdToTypeMap']


def sanitize_and_enforce_id_in_branch(branch_client: 'BranchClient', id: Union[str, int]):
    """Makes sure `id` is of the right type and exists in the Sedaro Branch associated with the `BranchClient`

    Args:
        branch_client (BranchClient): the `BranchClient` associated with the Sedaro Branch to check for the `id`
        id (Union[str, int]): `id` of the Sedaro Block to sanitize and check

    Raises:
        TypeError: if the `id` is not an integer or an integer string
        KeyError: if no corresponding Block exists in the Branch

    Returns:
        str: the integer string `id` of the Block
    """
    try:
        og_id = id
        id = str(id)
        assert id.isdigit()
    except:
        raise TypeError(
            f'The "id" argument must be a string of an integer or able to be coerced to such, not a "{type(og_id).__name__}".'
        )

    if id not in branch_client._block_id_to_type_map:
        raise KeyError(f'There is no Block with id "{id}" in this Branch.')

    return id


def import_if_exists(local_path: str):
    """Returns local import if exists, otherwise `None`

    Args:
        path (str): path to desired import

    Returns:
        import or `None`
    """
    if importlib.util.find_spec(local_path) is None:
        return None
    return importlib.import_module(local_path)


def get_class_from_module(module: ModuleType, target_class: str = None) -> type:
    """Gets a class from the passed in module.

    Args:
        module (ModuleType): python module (file)
        target_class (str, optional): name of the class to search for. Defaults to None.

    Raises:
        AttributeError: if no class is found to return

    Returns:
        type: If `target_class` not provided, returns first class defined in the module. If `target_class`
        is provided, returns class with a matching name (case-insensitive).
    """

    def filter_desired_class_from_file(kls):
        if not inspect.isclass(kls):
            return False
        if target_class is not None:
            return kls.__name__.casefold() == target_class.casefold()
        # if not searching for specific class name, return all that are defined in this module
        return kls.__module__ == module.__name__

    filtered_classes = inspect.getmembers(
        module,
        filter_desired_class_from_file
    )

    if not len(filtered_classes):
        if target_class is not None:
            err_msg = f'Module "{module.__name__}" has not attribute "{target_class}".'
        else:
            err_msg = f'There is no class defined in the module "{module.__name__}" to import.'
        raise AttributeError(err_msg)

    return filtered_classes[0][1]


@dataclass
class Wrap:
    body: dict


def temp_crud(sedaro_client, http_method):
    def outer(old_client_method):
        def wrapper(body=None, path_params=None, timeout=None):
            resource_path = getattr(
                sys.modules[old_client_method.__module__], 'path').value
            for k, v in path_params.items():
                resource_path = resource_path.replace('{%s}' % k, str(v))
            response = sedaro_client.call_api(
                resource_path=resource_path,
                method=http_method,
                body=json.dumps(body),
                timeout=timeout,
                headers={'Content-Type': 'application/json'}
            )
            parsed_response = parse_urllib_response(response)
            if response.status != 200:
                reason = parsed_response['error']['message'] if 'error' in parsed_response else 'An unknown error occurred.'
                raise SedaroApiException(
                    status=response.status, reason=reason)

            return Wrap(body=parsed_response)
        return wrapper
    return outer



from sedaro_base_client.paths.models_branches_branch_id_gnc_algorithms_attitude_determination_mekf_.post import ApiForpost


class ModelsBranchesBranchIdGncAlgorithmsAttitudeDeterminationMekf(
    ApiForpost,
):
    pass



import gzip
import json


from typing import Generator, Union, List
from pathlib import Path
from sedaro.results.block import SedaroBlockResult
from sedaro.results.utils import hfill, HFILL, ENGINE_EXPANSION


class SedaroAgentResult:

    def __init__(self, name: str, block_structures: dict, series: dict):
        '''Initialize a new agent result.

        Agent results are typically created through the .agent method of
        SedaroSimulationResult or the .from_file method of this class.
        '''
        self.__name = name
        self.__block_structures = block_structures
        self.__series = series
        self.__block_ids = sorted(set(
            block_id
            for module in self.__series
            for block_id in self.__series[module]['series']
            ),
            reverse=True
        )

    def __iter__(self) -> Generator:
        '''Iterate through blocks on this agent.'''
        return (self.block(id_) for id_ in self.__block_ids)

    def __contains__(self, id_: str) -> bool:
        '''Check if this agent result contains a certain block ID.'''
        return id_ in self.__block_ids

    @property
    def name(self) -> str:
        return self.__name

    @property
    def blocks(self) -> List[str]:
        return self.__block_ids

    def block(self, id_: str) -> SedaroBlockResult:
        '''Query results for a particular block by ID.'''
        id_ = str(id_)
        if id_ not in self.__block_ids:
            matching_id = [entry for entry in self.__block_ids if entry.startswith(id_)]
            if len(matching_id) == 1:
                id_ = matching_id[0]
            elif len(matching_id) == 0:
                raise ValueError(f"ID '{id_}' not found.")
            else:
                raise ValueError(f'Found multiple matching IDs for {id_}: {matching_id}.')

        block_data = {}
        for module in self.__series:
            if id_ in self.__series[module]['series']:
                if module not in block_data:
                    block_data[module] = {}
                block_data[module]['time'] = self.__series[module]['time']
                block_data[module]['series'] = self.__series[module]['series'][id_]
        block_structure = self.__block_structures[id_] if id_ != 'root' else id_
        return SedaroBlockResult(block_structure, block_data)

    def to_file(self, filename: Union[str, Path]) -> None:
        '''Save agent result to compressed JSON file.'''
        with gzip.open(filename, 'xt', encoding='UTF-8') as json_file:
            contents = {
                'name': self.__name,
                'block_structures': self.__block_structures,
                'series': self.__series,
            }
            json.dump(contents, json_file)
            print(f"💾 Successfully saved to {filename}")

    @classmethod
    def from_file(cls, filename: Union[str, Path]):
        '''Load agent result from compressed JSON file.'''
        with gzip.open(filename, 'rt', encoding='UTF-8') as json_file:
            contents = json.load(json_file)
            return cls(contents['name'], contents['block_structures'], contents['series'])

    def summarize(self) -> None:
        '''Summarize these results in the console.'''
        hfill()
        print(f"Agent Result Summary".center(HFILL))
        print(f"'{self.__name}'".center(HFILL))
        hfill()

        print("🧩 Simulated Modules")
        for module in self.__series:
            print(f'    • {ENGINE_EXPANSION[module]}')

        print("\n📦 Available Blocks")
        print('    ' + '-' * 68)
        print('    |' + 'id'.center(38) + 'name'.center(40-12) + '|')
        print('    ' + '-' * 68)
        for block_id in self.__block_ids:
            if block_id != 'root':
                block_name = self.__block_structures[block_id].get('name', None)
                block_id_col = f"{block_id[:36]}"
                if block_name is not None:
                    name_id_col = f'{block_name[:25]}'
                else:
                    name_id_col = f'<Unnamed Block>'
            else:
                block_id_col = f"root"
                name_id_col = ''
            print(f"    | {block_id_col:36s} | {name_id_col:25s} |")
        print('    ' + '-' * 68)

        no_data_blocks = len(self.__block_structures) - len(self.__block_ids)
        if no_data_blocks > 0:
            print(f"\n    {no_data_blocks} block(s) with no associated data")

        hfill()
        print("❓ Query block results with .block(<ID>) or .block(<PARTIAL_ID>)")



from sedaro_base_client.paths.models_branches_branch_id_power_solar_arrays_block_id.delete import ApiFordelete
from sedaro_base_client.paths.models_branches_branch_id_power_solar_arrays_block_id.patch import ApiForpatch


class ModelsBranchesBranchIdPowerSolarArraysBlockId(
    ApiFordelete,
    ApiForpatch,
):
    pass
